/******************************************************************************
 * arch/arm/mach-ep93xx/dma.c
 *
 * Support functions for the ep93xx internal DMA channels.
 * (see also Documentation/arm/ep93xx/dma.txt)
 *
 * Copyright (C) 2003  Cirrus Logic
 *
 * A large portion of this file is based on the dma api implemented by
 * Nicolas Pitre, dma-sa1100.c, copyrighted 2000.
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 ****************************************************************************/
/* #define DEBUG 1 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/interrupt.h>

#include <asm/system.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <asm/mach/dma.h>
#include <asm/hardware/vic.h>

/*****************************************************************************
 *
 * static global variables
 *
 ****************************************************************************/
ep93xx_dma_t ep93xx_dma_info[MAX_EP93XX_DMA_CHANNELS];

/* DMA Register Base addresses */
static unsigned int const ep93xx_dma_m2p_base[MAX_EP93XX_DMA_M2P_CHANNELS] =
{
	EP93XX_DMA_M2P_C0_TX_BASE,
	EP93XX_DMA_M2P_C1_RX_BASE,
	EP93XX_DMA_M2P_C2_TX_BASE,
	EP93XX_DMA_M2P_C3_RX_BASE,
	EP93XX_DMA_M2P_C4_TX_BASE,
	EP93XX_DMA_M2P_C5_RX_BASE,
	EP93XX_DMA_M2P_C6_TX_BASE,
	EP93XX_DMA_M2P_C7_RX_BASE,
	EP93XX_DMA_M2P_C8_TX_BASE,
	EP93XX_DMA_M2P_C9_RX_BASE
};

static unsigned int const ep93xx_dma_m2m_base[MAX_EP93XX_DMA_M2M_CHANNELS] =
{
	EP93XX_DMA_M2M_C0_BASE,
	EP93XX_DMA_M2M_C1_BASE
};

/*
 * lock used to protect the list of dma channels while searching for a free
 * channel during dma_request.
 */
static DEFINE_SPINLOCK(dma_list_lock);

/*****************************************************************************
 *
 * get_dma_channel_from_handle()
 *
 * If Handle is valid, returns the DMA channel # (0 to 9 for channels 1-10)
 * If Handle is not valid, returns -1.
 *
 ****************************************************************************/
static int
dma_get_channel_from_handle(int handle)
{
	int ch;

	/* Get the DMA channel # from the handle. */
	ch = (handle & DMA_HANDLE_SPECIFIER_MASK) >> 28;

	/* See if this is a valid handle. */
	if (ep93xx_dma_info[ch].last_valid_handle != handle) {
		pr_err("DMA ERROR - invalid handle 0x%x \n", handle);
		return -1;
	}

	/* See if this instance is still open */
	if (!ep93xx_dma_info[ch].ref_count)
		return -1;

	return ch;
}

static void dma_m2m_transfer_done(ep93xx_dma_t *dma)
{
	unsigned int ctl;
	unsigned int base = dma->reg_base;
	unsigned int bcr0, bcr1;

	pr_debug("1  ");

	__raw_writel(0, base + EP93XX_DMA_M2M_INTERRUPT_OFFSET);

	if (dma->total_buffers) {
		/* There is a small chance that the nfb handler added a buffer
		 * descriptor after the channel went into the DMA_STALL state.
		 * For software triggered M2M DMA, this will mean that the
		 * buffer has not been transfered. For IDE etc, the transfer may
		 * have already started. We must check for this */
		bcr0 = __raw_readl(dma->reg_base + EP93XX_DMA_M2M_BCR0_OFFSET);
		bcr1 = __raw_readl(dma->reg_base + EP93XX_DMA_M2M_BCR1_OFFSET);
		if (bcr0 || bcr1)
			return;

		/*
		 * The current_buffer has already been tranfered, so add the
		 * byte count to the total_bytes field.
		 */
		dma->total_bytes = dma->total_bytes +
			dma->buffer_queue[dma->current_buffer].size;

		/* Mark the current_buffer as used. */
		dma->buffer_queue[dma->current_buffer].used = 1;

		/* Increment the used buffer counter */
		dma->used_buffers++;

		pr_debug("#%d", dma->current_buffer);

		/* Increment the current_buffer */
		dma->current_buffer = (dma->current_buffer + 1) %
			MAX_EP93XX_DMA_BUFFERS;

		/*
		 * There's a chance we setup both buffer descriptors,
		 * but didn't service the NFB quickly enough, causing
		 * the channel to transfer both buffers, then enter the
		 * stall state.  So, we need to be able to process the
		 * second buffer.
		 */
		if ((dma->used_buffers + dma->new_buffers) < dma->total_buffers) {
			/*
			 * The current_buffer has already been
			 * tranferred, so add the byte count to the
			 * total_bytes field.
			 */
			dma->total_bytes = dma->total_bytes +
				dma->buffer_queue[dma->current_buffer].size;

			/* Mark the current_buffer as used. */
			dma->buffer_queue[dma->current_buffer].used = 1;

			/* Increment the used buffer counter */
			dma->used_buffers++;

			pr_debug("#%d", dma->current_buffer);

			/* Increment the current buffer pointer. */
			dma->current_buffer = (dma->current_buffer + 1) %
				MAX_EP93XX_DMA_BUFFERS;
		}

		/* check if there's a new buffer to transfer. */
		if (dma->new_buffers && dma->xfer_enable) {
			/*
			 * We have a new buffer to transfer so program in the
			 * buffer values.  Since a STALL interrupt was
			 * triggered, we program the buffer descriptor 0
			 *
			 * Set the SAR_BASE/DAR_BASE/BCR registers with values
			 * from the next buffer in the queue.
			 */
			__raw_writel(dma->buffer_queue[dma->current_buffer].source,
				     base + EP93XX_DMA_M2M_SAR_BASE0_OFFSET);

			__raw_writel(dma->buffer_queue[dma->current_buffer].dest,
				     base + EP93XX_DMA_M2M_DAR_BASE0_OFFSET);

			__raw_writel(dma->buffer_queue[dma->current_buffer].size,
				     base + EP93XX_DMA_M2M_BCR0_OFFSET);

			pr_debug("SAR_BASE0 - 0x%x\n",
				 dma->buffer_queue[dma->current_buffer].source);
			pr_debug("DAR_BASE0 - 0x%x\n",
				 dma->buffer_queue[dma->current_buffer].dest);
			pr_debug("BCR0 - 0x%x\n",
				 dma->buffer_queue[dma->current_buffer].size);

			/* Decrement the new buffer counter */
			dma->new_buffers--;

			/*
			 * If there's a second new buffer, we program the
			 * second buffer descriptor.
			 */
			if (dma->new_buffers) {
				__raw_writel(dma->buffer_queue[(dma->current_buffer + 1) %
							       MAX_EP93XX_DMA_BUFFERS].source,
					     base + EP93XX_DMA_M2M_SAR_BASE1_OFFSET);

				__raw_writel(dma->buffer_queue[(dma->current_buffer + 1) %
							       MAX_EP93XX_DMA_BUFFERS].dest,
					     base + EP93XX_DMA_M2M_DAR_BASE1_OFFSET);

				__raw_writel(dma->buffer_queue[(dma->current_buffer + 1) %
							       MAX_EP93XX_DMA_BUFFERS].size,
					     base + EP93XX_DMA_M2M_BCR1_OFFSET);

				ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
				ctl |= EP93XX_DMA_M2M_CONTROL_NFBINTEN;
				__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);

				dma->new_buffers--;
			}
		} else {

			/* No new buffers to transfer, so disable the channel. */
			ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
			ctl &= ~EP93XX_DMA_M2M_CONTROL_ENABLE;
			__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);

			/*
			 * Indicate that this channel is in the pause by
			 * starvation state by setting the pause bit to true.
			 */
			dma->pause = 1;
		}
	} else {
		/*
		 * No buffers to transfer, or old buffers to mark as used,
		 * so disable the channel
		 */
		ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		ctl &= ~EP93XX_DMA_M2M_CONTROL_ENABLE;
		__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);

		/* Must read the control register back after a write. */
		__raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);

		/*
		 * Indicate that this channel is in the pause by
		 * starvation state by setting the pause bit to true.
		 */
		dma->pause = 1;
	}
}

static void dump_m2m_register(unsigned int base)
{
	pr_debug("CONTROL   - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET));
	pr_debug("INTERRUPT - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2M_INTERRUPT_OFFSET));
	pr_debug("STATUS    - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2M_STATUS_OFFSET));
	pr_debug("BCR0      - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2M_BCR0_OFFSET));
	pr_debug("BCR1      - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2M_BCR1_OFFSET));
	pr_debug("SAR_BASE0 - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2M_SAR_BASE0_OFFSET));
	pr_debug("SAR_CUR0  - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2M_SAR_CURRENT0_OFFSET));
	pr_debug("SAR_CUR1  - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2M_SAR_CURRENT1_OFFSET));
	pr_debug("DAR_BASE0 - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2M_DAR_BASE0_OFFSET));
	pr_debug("DAR_BASE1 - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2M_DAR_BASE1_OFFSET));
	pr_debug("DAR_CUR0  - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2M_DAR_CURRENT0_OFFSET));
	pr_debug("DAR_CUR1  - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2M_DAR_CURRENT1_OFFSET));
}

static void dump_m2p_register(unsigned int base)
{
	pr_debug("CONTROL   - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2P_CONTROL_OFFSET));
	pr_debug("INTERRUPT - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2P_INTERRUPT_OFFSET));
	pr_debug("PPALLOC   - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2P_PPALLOC_OFFSET));
	pr_debug("STATUS    - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2P_STATUS_OFFSET));
	pr_debug("REMAIN    - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2P_REMAIN_OFFSET));
	pr_debug("MAXCNT0   - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2P_MAXCNT0_OFFSET));
	pr_debug("BASE0     - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2P_BASE0_OFFSET));
	pr_debug("CURRENT0  - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2P_CURRENT0_OFFSET));
	pr_debug("MAXCNT1   - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2P_MAXCNT1_OFFSET));
	pr_debug("BASE1     - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2P_BASE1_OFFSET));
	pr_debug("CURRENT1  - 0x%x \n",
		 __raw_readl(base + EP93XX_DMA_M2P_CURRENT1_OFFSET));
}

static void dump_dma_info(ep93xx_dma_t *dma)
{
	int i;

	pr_debug("Buffer	buf_id		source	size	last	used\n");
	for (i = 0; i < 32; i ++)
		pr_debug("%d	0x%x	0x%x	0x%x	%d	%d\n",
			 i, dma->buffer_queue[i].buf_id, dma->buffer_queue[i].source,
			 dma->buffer_queue[i].size,
			 dma->buffer_queue[i].last, dma->buffer_queue[i].used);
	pr_debug("pause	0x%x	0x%x	0x%x	%d	%d\n",
		 dma->pause_buf.buf_id, dma->pause_buf.source, dma->pause_buf.size,
		 dma->pause_buf.last, dma->pause_buf.used);

	pr_debug("Pause          - %d \n", dma->pause);
	pr_debug("xfer_enable    - %d \n", dma->xfer_enable);
	pr_debug("total bytes    - 0x%x \n", dma->total_bytes);
	pr_debug("total buffer   - %d \n", dma->total_buffers);
	pr_debug("new buffers    - %d \n", dma->new_buffers);
	pr_debug("current buffer - %d \n", dma->current_buffer);
	pr_debug("last buffer    - %d \n", dma->last_buffer);
	pr_debug("used buffers   - %d \n", dma->used_buffers);
	pr_debug("callback addr  - 0x%p \n", dma->callback);
}

static void dma_m2m_next_frame_buffer(ep93xx_dma_t *dma)
{
	unsigned int ctl, status;
	unsigned int base = dma->reg_base;

	pr_debug("5  ");

	if (dma->total_buffers) {
		pr_debug("6  ");
		/*
		 * The iCurrentBuffer has already been transfered.  so add the
		 * byte count from the current buffer to the total byte count.
		 */
		dma->total_bytes = dma->total_bytes +
			dma->buffer_queue[dma->current_buffer].size;

		/* Mark the Current Buffer as used. */
		dma->buffer_queue[dma->current_buffer].used = 1;

		/* Increment the used buffer counter */
		dma->used_buffers++;

		pr_debug("#%d", dma->current_buffer);

		if ((dma->buffer_queue[(dma->current_buffer + 1) %
				       MAX_EP93XX_DMA_BUFFERS].last) ||
		    (dma->new_buffers == 0) || (dma->xfer_enable == 0)) {
			pr_debug("7  ");

			/*
			 * This is the last Buffer in this transaction, so
			 * disable the NFB interrupt.  We shouldn't get an NFB
			 * int when the FSM moves to the ON state where it
			 * would typically get the NFB int indicating a new
			 * buffer can be programmed.  Instead, once in the ON
			 * state, the DMA will just proceed to complete the
			 * transfer of the current buffer, move the FSB
			 * directly to the STALL state where a STALL interrupt
			 * will be generated.
			 */
			ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
			ctl &= ~EP93XX_DMA_M2M_CONTROL_NFBINTEN;
			__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);

			/*
			 * The current buffer has been transferred, so
			 * increment the current buffer counter to reflect
			 * this.
			 */
			dma->current_buffer = (dma->current_buffer + 1) %
				MAX_EP93XX_DMA_BUFFERS;

			pr_debug("End of NFB handling. \n");

			dump_m2m_register(base);
			dump_dma_info(dma);
		} else if (dma->new_buffers) {
			pr_debug("8  ");
			/*
			 * We have a new buffer, so increment the current
			 * buffer to point to the next buffer, which is already
			 * programmed into the DMA. Next time around, it'll be
			 * pointing to the current buffer.
			 */
			dma->current_buffer = (dma->current_buffer + 1) %
				MAX_EP93XX_DMA_BUFFERS;

			/*
			 * We know we have a new buffer to program as the next
			 * buffer, so check which set of SAR_BASE/DAR_BASE/BCR
			 * registers to program.
			 */
			status = __raw_readl(base + EP93XX_DMA_M2M_STATUS_OFFSET);
			if (status & EP93XX_DMA_M2M_STATUS_NB) {
				/*
				 * Set the SAR_BASE1/DAR_BASE1/BCR1 registers
				 * with values from the next buffer in the
				 * queue.
				 */
				__raw_writel(dma->buffer_queue[(dma->current_buffer + 1) %
							       MAX_EP93XX_DMA_BUFFERS].source,
					     base + EP93XX_DMA_M2M_SAR_BASE1_OFFSET);

				__raw_writel(dma->buffer_queue[(dma->current_buffer + 1) %
							       MAX_EP93XX_DMA_BUFFERS].dest,
					     base + EP93XX_DMA_M2M_DAR_BASE1_OFFSET);

				__raw_writel(dma->buffer_queue[(dma->current_buffer + 1) %
							       MAX_EP93XX_DMA_BUFFERS].size,
					     base + EP93XX_DMA_M2M_BCR1_OFFSET);
			} else {
				/*
				 * Set the SAR_BASE0/DAR_BASE0/BCR0 registers
				 * with values from the next buffer in the
				 * queue.
				 */
				__raw_writel(dma->buffer_queue[(dma->current_buffer + 1) %
							       MAX_EP93XX_DMA_BUFFERS].source,
					     base + EP93XX_DMA_M2M_SAR_BASE0_OFFSET);

				__raw_writel(dma->buffer_queue[(dma->current_buffer + 1) %
							       MAX_EP93XX_DMA_BUFFERS].dest,
					     base + EP93XX_DMA_M2M_DAR_BASE0_OFFSET);

				__raw_writel(dma->buffer_queue[(dma->current_buffer + 1) %
							       MAX_EP93XX_DMA_BUFFERS].size,
					     base + EP93XX_DMA_M2M_BCR0_OFFSET);
			}

			/* Decrement the new buffers counter */
			dma->new_buffers--;
		}
	} else {
		/*
		 * Total number of buffers is 0 - really we should never get
		 * here, but just in case.
		 */
		pr_debug("9 \n");

		/* No new buffers to transfer, so Disable the channel */
		ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		ctl &= ~EP93XX_DMA_M2M_CONTROL_ENABLE;
		__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);

		/* Indicate that the channel is paused by starvation. */
		dma->pause = 1;
	}
}

/*****************************************************************************
 *
 * dma_m2m_irq_handler
 *
 ****************************************************************************/
static irqreturn_t
dma_m2m_irq_handler(int irq, void *data)
{
	ep93xx_dma_t *dma = (ep93xx_dma_t *)data;
	unsigned int base = dma->reg_base;
	ep93xx_dma_int_t dma_int = UNDEF_INT;
	int status;

	/* Determine what kind of dma interrupt this is. */
	status = __raw_readl(base + EP93XX_DMA_M2M_INTERRUPT_OFFSET);
	if (status & EP93XX_DMA_M2M_INTERRUPT_DONEINT)
		dma_int = DONE; /* we're done with a requested dma */
	else if (status & EP93XX_DMA_M2M_INTERRUPT_NFBINT)
		dma_int = NFB;  /* we're done with one dma buffer */

	pr_debug("IRQ: b=%#x st=%#x\n", (int)dma->current_buffer, dma_int);

	switch (dma_int) {
		/*
		 * Next Frame Buffer Interrupt.  If there's a new buffer program it
		 * Check if this is the last buffer in the transfer,
		 * and if it is, disable the NFB int to prevent being
		 * interrupted for another buffer when we know there won't be
		 * another.
		 */
	case NFB:
		dma_m2m_next_frame_buffer(dma);
		break;
		/*
		 * Done interrupt generated, indicating that the transfer
		 * is complete.
		 */
	case DONE:
		dma_m2m_transfer_done(dma);
		break;

	default:
		break;
	}

	if ((dma_int != UNDEF_INT) && dma->callback)
		dma->callback(dma_int, dma->device, dma->user_data);

	return IRQ_HANDLED;
}

/*****************************************************************************
 *
 * dma_m2p_irq_handler
 *
 ****************************************************************************/
static irqreturn_t
dma_m2p_irq_handler(int irq, void *data)
{
	ep93xx_dma_t *dma = (ep93xx_dma_t *)data;
	unsigned int base = dma->reg_base;
	ep93xx_dma_int_t dma_int = UNDEF_INT;
	unsigned int control, interrupt, status;

	/* Determine what kind of dma interrupt this is. */
	status = __raw_readl(base + EP93XX_DMA_M2P_INTERRUPT_OFFSET);
	if (status & EP93XX_DMA_M2P_INTERRUPT_STALLINT)
		dma_int = STALL;
	else if (status & EP93XX_DMA_M2P_INTERRUPT_NFBINT)
		dma_int = NFB;
	else if (status & EP93XX_DMA_M2P_INTERRUPT_CHERRORINT)
		dma_int = CHERROR;

	/*
	 * Stall Interrupt: The Channel is stalled, meaning nothing is
	 * programmed to transfer right now.  So, we're back to the
	 * beginnning.  If there's a buffer to transfer, program it into
	 * max and base 0 registers.
	 */
	if (dma_int == STALL) {
		pr_debug("1  ");

		if (dma->total_buffers) {
			/*
			 * The current_buffer has already been tranfered, so
			 * add the byte count to the total_bytes field.
			 */
			dma->total_bytes = dma->total_bytes +
				dma->buffer_queue[dma->current_buffer].size;

			/* Mark the current_buffer as used. */
			dma->buffer_queue[dma->current_buffer].used = 1;

			/* Increment the used buffer counter */
			dma->used_buffers++;

			pr_debug("#%d", dma->current_buffer);

			/* Increment the current_buffer */
			dma->current_buffer = (dma->current_buffer + 1) %
				MAX_EP93XX_DMA_BUFFERS;

			/* check if there's a new buffer to transfer. */
			if (dma->new_buffers && dma->xfer_enable) {
				/*
				 * We have a new buffer to transfer so program
				 * in the buffer values.  Since a STALL
				 * interrupt was triggered, we program the
				 * base0 and maxcnt0
				 *
				 * Set the MAXCNT0 register with the buffer
				 * size
				 */
				__raw_writel(dma->buffer_queue[dma->current_buffer].size,
					     base + EP93XX_DMA_M2P_MAXCNT0_OFFSET );

				/*
				 * Set the BASE0 register with the buffer base
				 * address
				 */
				__raw_writel(dma->buffer_queue[dma->current_buffer].source,
					     base + EP93XX_DMA_M2P_BASE0_OFFSET );

				/* Decrement the new buffer counter */
				dma->new_buffers--;

				if (dma->new_buffers) {
					pr_debug("A  ");
					/*
					 * Set the MAXCNT1 register with the
					 * buffer size
					 */
					__raw_writel(dma->buffer_queue[(dma->current_buffer + 1) %
								       MAX_EP93XX_DMA_BUFFERS].size,
						     base + EP93XX_DMA_M2P_MAXCNT1_OFFSET);

					/*
					 * Set the BASE1 register with the
					 * buffer base address
					 */
					__raw_writel(dma->buffer_queue[dma->current_buffer + 1 %
								       MAX_EP93XX_DMA_BUFFERS].source,
						     base + EP93XX_DMA_M2P_BASE1_OFFSET);

					/* Decrement the new buffer counter */
					dma->new_buffers--;

					/* Enable the NFB Interrupt. */
					control = __raw_readl(base + EP93XX_DMA_M2P_CONTROL_OFFSET);
					control |= EP93XX_DMA_M2P_CONTROL_NFBINT;
					__raw_writel(control, base + EP93XX_DMA_M2P_CONTROL_OFFSET);
				}
			} else {
				/* No new buffers. */
				pr_debug("2 \n");

				/*
				 * There's a chance we setup both buffer descriptors, but
				 * didn't service the NFB quickly enough, causing the channel
				 * to transfer both buffers, then enter the stall state.
				 * So, we need to be able to process the second buffer.
				 */
				if ((dma->used_buffers + dma->new_buffers) < dma->total_buffers) {
					pr_debug("3 ");

					/*
					 * The current_buffer has already been tranfered, so add the
					 * byte count to the total_bytes field.
					 */
					dma->total_bytes = dma->total_bytes +
						dma->buffer_queue[dma->current_buffer].size;

					/* Mark the current_buffer as used. */
					dma->buffer_queue[dma->current_buffer].used = 1;

					/* Increment the used buffer counter */
					dma->used_buffers++;

					pr_debug("#%d", dma->current_buffer);

					/* Increment the current buffer pointer. */
					dma->current_buffer = (dma->current_buffer + 1) %
						MAX_EP93XX_DMA_BUFFERS;
				}

				/*
				 * No new buffers to transfer, so disable the channel.
				 */
				control = __raw_readl(base + EP93XX_DMA_M2P_CONTROL_OFFSET);
				control &= ~EP93XX_DMA_M2P_CONTROL_ENABLE;
				__raw_writel(control, base + EP93XX_DMA_M2P_CONTROL_OFFSET);

				/*
				 * Indicate that this channel is in the pause by starvation
				 * state by setting the pause bit to true.
				 */
				dma->pause = 1;

				dump_m2p_register(base);
				dump_dma_info(dma);
			}
		} else {
			/*
			 * No buffers to transfer, or old buffers to mark as used,
			 * so Disable the channel
			 */
			control = __raw_readl(base + EP93XX_DMA_M2P_CONTROL_OFFSET);
			control &= ~EP93XX_DMA_M2P_CONTROL_ENABLE;
			__raw_writel(control, base + EP93XX_DMA_M2P_CONTROL_OFFSET);

			/* Must read the control register back after a write. */
			__raw_readl(base + EP93XX_DMA_M2P_CONTROL_OFFSET);

			/*
			 * Indicate that this channel is in the pause by
			 * starvation state by setting the pause bit to true.
			 */
			dma->pause = 1;
		}
	}

	/*
	 * Next Frame Buffer Interrupt.  If there's a new buffer program it
	 * Check if this is the last buffer in the transfer,
	 * and if it is, disable the NFB int to prevent being
	 * interrupted for another buffer when we know there won't be
	 * another.
	 */
	if (dma_int == NFB) {
		pr_debug("5  ");

		if (dma->total_buffers) {
			pr_debug("6  ");
			/*
			 * The iCurrentBuffer has already been transfered.  so add the
			 * byte count from the current buffer to the total byte count.
			 */
			dma->total_bytes = dma->total_bytes +
				dma->buffer_queue[dma->current_buffer].size;

			/* Mark the Current Buffer as used. */
			dma->buffer_queue[dma->current_buffer].used = 1;

			/* Increment the used buffer counter */
			dma->used_buffers++;

			pr_debug("#%d", dma->current_buffer);

			if ((dma->buffer_queue[(dma->current_buffer + 1) %
					       MAX_EP93XX_DMA_BUFFERS].last) ||
			    (dma->new_buffers == 0) || (dma->xfer_enable == 0)) {
				pr_debug("7  ");

				/*
				 * This is the last Buffer in this transaction, so disable
				 * the NFB interrupt.  We shouldn't get an NFB int when the
				 * FSM moves to the ON state where it would typically get the
				 * NFB int indicating a new buffer can be programmed.
				 * Instead, once in the ON state, the DMA will just proceed
				 * to complet the transfer of the current buffer, move the
				 * FSB directly to the STALL state where a STALL interrupt
				 * will be generated.
				 */
				control = __raw_readl(base + EP93XX_DMA_M2P_CONTROL_OFFSET);
				control &= ~EP93XX_DMA_M2P_CONTROL_NFBINT;
				__raw_writel(control, base + EP93XX_DMA_M2P_CONTROL_OFFSET);

				/*
				 * The current buffer has been transferred, so increment
				 * the current buffer counter to reflect this.
				 */
				dma->current_buffer = (dma->current_buffer + 1) %
					MAX_EP93XX_DMA_BUFFERS;

				pr_debug("End of NFB handling. \n");

				dump_m2p_register(base);
				dump_dma_info(dma);
			} else if (dma->new_buffers) {
				pr_debug("8  ");
				/*
				 * we have a new buffer, so increment the current buffer to
				 * point to the next buffer, which is already programmed into
				 * the DMA. Next time around, it'll be pointing to the
				 * current buffer.
				 */
				dma->current_buffer = (dma->current_buffer + 1) %
					MAX_EP93XX_DMA_BUFFERS;

				/*
				 * we know we have a new buffer to program as the next
				 * buffer, so check which set of MAXCNT and BASE registers
				 * to program.
				 */
				if (__raw_readl(base + EP93XX_DMA_M2P_STATUS_OFFSET) &
				    EP93XX_DMA_M2P_STATUS_NEXTBUFFER) {
					/*
					 * Set the MAXCNT1 register with the buffer size
					 */
					__raw_writel(dma->buffer_queue[(dma->current_buffer +1) %
								       MAX_EP93XX_DMA_BUFFERS].size,
						     base + EP93XX_DMA_M2P_MAXCNT1_OFFSET);

					/*
					 * Set the BASE1 register with the buffer base address
					 */
					__raw_writel(dma->buffer_queue[(dma->current_buffer + 1) %
								       MAX_EP93XX_DMA_BUFFERS].source,
						     base + EP93XX_DMA_M2P_BASE1_OFFSET);
				} else {
					/*
					 * Set the MAXCNT0 register with the buffer size
					 */
					__raw_writel(dma->buffer_queue[(dma->current_buffer + 1) %
								       MAX_EP93XX_DMA_BUFFERS].size,
						     base + EP93XX_DMA_M2P_MAXCNT0_OFFSET);

					/*
					 * Set the BASE0 register with the buffer base address
					 */
					__raw_writel(dma->buffer_queue[(dma->current_buffer + 1) %
								       MAX_EP93XX_DMA_BUFFERS].source,
						     base + EP93XX_DMA_M2P_BASE0_OFFSET);
				}

				/* Decrement the new buffers counter */
				dma->new_buffers--;
			}
		} else {
			/*
			 * Total number of buffers is 0 - really we should never get here,
			 * but just in case.
			 */
			pr_debug("9 \n");

			/* No new buffers to transfer, so Disable the channel */
			control = __raw_readl(base + EP93XX_DMA_M2P_CONTROL_OFFSET);
			control &= ~EP93XX_DMA_M2P_CONTROL_ENABLE;
			__raw_writel(control, base + EP93XX_DMA_M2P_CONTROL_OFFSET);
		}
	}

	/*
	 * Channel Error Interrupt, or perhipheral interrupt, specific to the
	 * memory to/from peripheral channels.
	 */
	if (dma_int == CHERROR) {
		/*
		 * just clear the interrupt, it's really up to the peripheral
		 * driver to determine if any further action is necessary.
		 */
		interrupt = __raw_readl(base + EP93XX_DMA_M2P_INTERRUPT_OFFSET);
		interrupt &= ~EP93XX_DMA_M2P_INTERRUPT_CHERRORINT;
		__raw_writel(interrupt, base + EP93XX_DMA_M2P_INTERRUPT_OFFSET);
	}

	/*
	 * Make sure the interrupt was valid, and if it was, then check
	 * if a callback function was installed for this DMA channel.  If a
	 * callback was installed call it.
	 */
	if ((dma_int != UNDEF_INT) && dma->callback)
		dma->callback(dma_int, dma->device, dma->user_data);

	return IRQ_HANDLED;
}

/*****************************************************************************
 *
 * ep9312_dma_open_m2p(int device)
 *
 * Description: This function will attempt to open a M2P/P2M DMA channel.
 *		If the open is successful, the channel number is returned,
 *		otherwise a negative number is returned.
 *
 * Parameters:
 * device:	 device for which the dma channel is requested.
 *
 ****************************************************************************/
static int
dma_open_m2p(int device)
{
	int ch = -1;
	unsigned int i;
	unsigned int base;
	unsigned int clkctl;
	unsigned long flags;

	pr_debug("DMA Open M2P with hw dev %d\n", device);

	/* Lock the dma channel list. */
	spin_lock_irqsave(&dma_list_lock, flags);

	pr_debug("1\n");

	/* Verify that the device requesting DMA isn't already using a DMA channel */
	if (device >= 10)
		i = 1;	/* Rx transfer requested */
	else
		i = 0;	/* Tx transfer requested */

	for (; i < MAX_EP93XX_DMA_M2P_CHANNELS; i += 2)
		/*
		 * Before checking for a matching device, check that the
		 * channel is in use, otherwise the device field is
		 * invalid.
		 */
		if (ep93xx_dma_info[i].ref_count)
			if (device == ep93xx_dma_info[i].device) {
				pr_err("DMA Open M2P - Error\n");
				return -1;
			}

	/*
	 * Get a DMA channel instance for the given hardware device.
	 * If this is a TX look for even numbered channels, else look for
	 * odd numbered channels
	 */
	if (device >= 10)
		i = 1;	/* Rx transfer requested */
	else
		i = 0;	/* Tx transfer requested */

	for (; i < MAX_EP93XX_DMA_M2P_CHANNELS; i += 2)
		if (!ep93xx_dma_info[i].ref_count) {
			/* Capture the channel and increment the reference count. */
			ch = i;
			ep93xx_dma_info[ch].ref_count++;
			break;
		}

	/* Unlock the dma channel list. */
	spin_unlock_irqrestore(&dma_list_lock, flags);

	/* See if we got a valid channel. */
	if (ch < 0)
		return -1;

	/* Point regs to the correct dma channel register base. */
	base = ep93xx_dma_info[ch].reg_base;

	/*
	 * Turn on the clock for the specified DMA channel
	 * TODO: need to use the correct register name for the
	 * power control register.
	 */
	clkctl = __raw_readl(EP93XX_SYSCON_CLOCK_CONTROL);
	switch (ch) {
	case 0:
		clkctl |= EP93XX_SYSCON_CLOCK_DMA_M2P_C0;
		break;
	case 1:
		clkctl |= EP93XX_SYSCON_CLOCK_DMA_M2P_C1;
		break;
	case 2:
		clkctl |= EP93XX_SYSCON_CLOCK_DMA_M2P_C2;
		break;
	case 3:
		clkctl |= EP93XX_SYSCON_CLOCK_DMA_M2P_C3;
		break;
	case 4:
		clkctl |= EP93XX_SYSCON_CLOCK_DMA_M2P_C4;
		break;
	case 5:
		clkctl |= EP93XX_SYSCON_CLOCK_DMA_M2P_C5;
		break;
	case 6:
		clkctl |= EP93XX_SYSCON_CLOCK_DMA_M2P_C6;
		break;
	case 7:
		clkctl |= EP93XX_SYSCON_CLOCK_DMA_M2P_C7;
		break;
	case 8:
		clkctl |= EP93XX_SYSCON_CLOCK_DMA_M2P_C8;
		break;
	case 9:
		clkctl |= EP93XX_SYSCON_CLOCK_DMA_M2P_C9;
		break;
	default:
		return -1;
	}
	__raw_writel(clkctl, EP93XX_SYSCON_CLOCK_CONTROL);

	/* Clear out the control register before any further setup. */
	__raw_writel(0, base + EP93XX_DMA_M2P_CONTROL_OFFSET);

	/* Setup the peripheral port value in the DMA channel registers. */
	if (device < 10)
		__raw_writel((unsigned int)device,
			     base + EP93XX_DMA_M2P_PPALLOC_OFFSET);
	else
		__raw_writel((unsigned int)(device - 10),
			     base + EP93XX_DMA_M2P_PPALLOC_OFFSET);

	/* Let's hold on to the value of the Hw device for comparison later. */
	ep93xx_dma_info[ch].device = device;

	/* Success. */
	return ch;
}

/*****************************************************************************
 *
 * dma_open_m2m(int device)
 *
 * Description: This function will attempt to open a M2M DMA channel.
 *		If the open is successful, the channel number is returned,
 *		otherwise a negative number is returned.
 *
 * Parameters:
 *  device:	device for which the dma channel is requested.
 *
 ****************************************************************************/
static int
dma_open_m2m(int device)
{
	int ch = -1;
	unsigned int i;
	unsigned int base;
	unsigned int clkctl, ctl;
	unsigned long flags;

	pr_debug("DMA Open M2M with hw dev %d\n", device);

	/* Lock the dma channel list. */
	spin_lock_irqsave(&dma_list_lock, flags);

	pr_debug("1\n");

	/*
	 * Check if this device is already allocated a channel.
	 * TODO: can one M2M device be allocated multiple channels?
	 */
	for (i = DMA_MEMORY; i < UNDEF; i++)
		/*
		 * Before checking for a matching device, check that the
		 * channel is in use, otherwise the device field is
		 * invalid.
		 */
		if (ep93xx_dma_info[i].ref_count)
			if (device == ep93xx_dma_info[i].device) {
				pr_err("Error - dma_open_m2m - already allocated "
				       "channel\n");

				/* Unlock the dma channel list. */
				spin_unlock_irqrestore(&dma_list_lock, flags);

				/* Fail. */
				return -1;
			}

	/* Get a DMA channel instance for the given hardware device. */
	for (i = MAX_EP93XX_DMA_M2P_CHANNELS; i < MAX_EP93XX_DMA_CHANNELS; i++)
		if (!ep93xx_dma_info[i].ref_count) {
			/* Capture the channel and increment the reference count. */
			ch = i;
			ep93xx_dma_info[ch].ref_count++;
			break;
		}

	/* Unlock the dma channel list. */
	spin_unlock_irqrestore(&dma_list_lock, flags);

	/* See if we got a valid channel. */
	if (ch < 0)
		return -1;

	/* Point regs to the correct dma channel register base. */
	base = ep93xx_dma_info[ch].reg_base;

	/*
	 * Turn on the clock for the specified DMA channel
	 * TODO: need to use the correct register name for the
	 * power control register.
	 */
	clkctl = __raw_readl(EP93XX_SYSCON_CLOCK_CONTROL);
	switch (ch) {
	case 10:
		clkctl |= EP93XX_SYSCON_CLOCK_DMA_M2M_C0;
		break;

	case 11:
		clkctl |= EP93XX_SYSCON_CLOCK_DMA_M2M_C1;
		break;

	default:
		return -1;
	}
	__raw_writel(clkctl, EP93XX_SYSCON_CLOCK_CONTROL);

	pr_debug("DMA Open - power control: 0x%x \n",
		 __raw_readl(EP93XX_SYSCON_CLOCK_CONTROL));

	/* Clear out the control register before any further setup. */
	__raw_writel(0, base + EP93XX_DMA_M2M_CONTROL_OFFSET);

	/*
	 * Setup the transfer mode and the request source selection within
	 * the DMA M2M channel registers.
	 */
	switch (device) {
	case DMA_MEMORY:
		/* Clear TM field, set RSS field to 0 */
		ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		ctl &= ~(EP93XX_DMA_M2M_CONTROL_TM_MASK |
			 EP93XX_DMA_M2M_CONTROL_RSS_MASK);
		__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		break;
	case DMA_IDE:
		/* Set RSS field to 3, Set NO_HDSK, Set PW field to 1 */
		ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		ctl &= ~(EP93XX_DMA_M2M_CONTROL_RSS_MASK |
			 EP93XX_DMA_M2M_CONTROL_PW_MASK);
		ctl |= (EP93XX_DMA_M2M_CONTROL_RSS_IDE <<
			EP93XX_DMA_M2M_CONTROL_RSS_SHIFT) |
			EP93XX_DMA_M2M_CONTROL_NO_HDSK |
			(EP93XX_DMA_M2M_CONTROL_PW_32 <<
			 EP93XX_DMA_M2M_CONTROL_PW_SHIFT);
		ctl &= ~(EP93XX_DMA_M2M_CONTROL_ETDP_MASK);
		ctl &= ~(EP93XX_DMA_M2M_CONTROL_DACKP);
		ctl &= ~(EP93XX_DMA_M2M_CONTROL_DREQP_MASK);

		__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		__raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		break;
	case DMARx_SSP:
		/* Set RSS field to 1, Set NO_HDSK, Set TM field to 2 */
		ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		ctl &= ~(EP93XX_DMA_M2M_CONTROL_RSS_MASK |
			 EP93XX_DMA_M2M_CONTROL_TM_MASK);
		ctl |= (EP93XX_DMA_M2M_CONTROL_RSS_SSPRX <<
			EP93XX_DMA_M2M_CONTROL_RSS_SHIFT) |
			EP93XX_DMA_M2M_CONTROL_NO_HDSK |
			(EP93XX_DMA_M2M_CONTROL_TM_HARD_P2M <<
			 EP93XX_DMA_M2M_CONTROL_TM_SHIFT);
		__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		break;
	case DMATx_SSP:
		/* Set RSS field to 2, Set NO_HDSK, Set TM field to 1 */
		ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		ctl &= ~(EP93XX_DMA_M2M_CONTROL_RSS_MASK |
			 EP93XX_DMA_M2M_CONTROL_TM_MASK);
		ctl |= (EP93XX_DMA_M2M_CONTROL_RSS_SSPTX <<
			EP93XX_DMA_M2M_CONTROL_RSS_SHIFT) |
			EP93XX_DMA_M2M_CONTROL_NO_HDSK |
			(EP93XX_DMA_M2M_CONTROL_TM_HARD_M2P <<
			 EP93XX_DMA_M2M_CONTROL_TM_SHIFT);
		__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		break;
	case DMATx_EXT_DREQ:
		/* Set TM field to 2, set RSS field to 0 */
		ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		ctl &= ~(EP93XX_DMA_M2M_CONTROL_RSS_MASK |
			 EP93XX_DMA_M2M_CONTROL_TM_MASK);
		ctl |= EP93XX_DMA_M2M_CONTROL_TM_HARD_M2P <<
			EP93XX_DMA_M2M_CONTROL_TM_SHIFT;
		__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		break;

	case DMARx_EXT_DREQ:
		/* Set TM field to 2, set RSS field to 0 */
		ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		ctl &= ~(EP93XX_DMA_M2M_CONTROL_RSS_MASK |
			 EP93XX_DMA_M2M_CONTROL_TM_MASK);
		ctl |= EP93XX_DMA_M2M_CONTROL_TM_HARD_P2M <<
			EP93XX_DMA_M2M_CONTROL_TM_SHIFT;
		__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		break;
	default:
		return -1;
	}

	/* Let's hold on to the value of the Hw device for comparison later. */
	ep93xx_dma_info[ch].device = device;

	/* Success. */
	return ch;
}

/*****************************************************************************
 *
 * int dma_config_m2m(ep93xx_dma_t * dma, unsigned int flags_m2m,
 *			dma_callback callback, unsigned int user_data)
 *
 * Description: Configure the DMA channel and install a callback function.
 *		This function will have to be called for every transfer
 *
 * dma:		Pointer to the dma instance data for the M2M channel to
 *		configure.
 * flags_m2m:   Flags used to configure an M2M dma channel and determine
 *		if a callback function and user_data information are included
 *		in this call.
 * callback:	function pointer which is called near the end of the
 *		dma channel's irq handler.
 * user_data:   defined by the calling driver.
 *
 ****************************************************************************/
static int
dma_config_m2m(ep93xx_dma_t * dma, unsigned int flags_m2m,
	       dma_callback callback, unsigned int user_data)
{
	unsigned long flags;
	unsigned int base, ctl;

	/*
	 * Make sure the channel is disabled before configuring the channel.
	 *
	 * TODO: Is this correct??   Making a big change here...
	 */
	/* if (!dma->pause || (!dma->pause && dma->xfer_enable)) */
	if (dma->xfer_enable) {
		/* DMA channel is not paused, so we can't configure it. */
		pr_err("DMA channel not paused, so can't configure!\n");
		return -1;
	}

	/* Mask interrupts. */
	local_irq_save(flags);

	/* Setup a pointer into the dma channel's register set. */
	base = dma->reg_base;

	ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
	__raw_writel(0, base + EP93XX_DMA_M2M_CONTROL_OFFSET);
	__raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
	__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);

	/* By default we disable the stall interrupt. */
	ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
	ctl &= ~EP93XX_DMA_M2M_CONTROL_STALLINTEN;
	__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);

	/* By default we disable the done interrupt. */
	ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
	ctl &= ~EP93XX_DMA_M2M_CONTROL_DONEINTEN;
	__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);

	/*
	 * Set up the transfer control fields based on values passed in
	 * the flags_m2m field.
	 */
	ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);

	if (flags_m2m & DESTINATION_HOLD)
		ctl |= EP93XX_DMA_M2M_CONTROL_DAH;
	else
		ctl &= ~EP93XX_DMA_M2M_CONTROL_DAH;

	if (flags_m2m & SOURCE_HOLD)
		ctl |= EP93XX_DMA_M2M_CONTROL_SAH;
	else
		ctl &= ~EP93XX_DMA_M2M_CONTROL_SAH;

	ctl &= ~EP93XX_DMA_M2M_CONTROL_TM_MASK;
	ctl |= (((flags_m2m & TRANSFER_MODE_MASK) >> TRANSFER_MODE_SHIFT) <<
		EP93XX_DMA_M2M_CONTROL_TM_SHIFT) &
		EP93XX_DMA_M2M_CONTROL_TM_MASK;

	ctl &= ~EP93XX_DMA_M2M_CONTROL_PWSC_MASK;
	ctl |= (((flags_m2m & WAIT_STATES_MASK) >> WAIT_STATES_SHIFT) <<
		EP93XX_DMA_M2M_CONTROL_PWSC_SHIFT) &
		EP93XX_DMA_M2M_CONTROL_PWSC_MASK;

	__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);
	__raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);

	/* Save the callback function in the dma instance for this channel. */
	dma->callback = callback;

	/* Save the user data in the the dma instance for this channel. */
	dma->user_data = user_data;

	/*
	 * Put the dma instance into the pause state by setting the
	 * pause bit to true.
	 */
	dma->pause = 1;

	local_irq_restore(flags);

	/* Success. */
	return 0;
}

/*****************************************************************************
 *
 * int dma_start(int handle, unsigned int channels, unsigned int * handles)
 *
 * Description: Initiate a transfer on up to 3 channels.
 *
 * handle:	handle for the channel to initiate transfer on.
 * channels:	number of channels to initiate transfers on.
 * handles:	pointer to an array of handles, one for each channel which
 *		is to be started.
 *
 ****************************************************************************/
static int
dma_start_m2m(int channel, ep93xx_dma_t *dma)
{
	unsigned long flags;
	unsigned int base = dma->reg_base;
	unsigned int ctl;

	/* Mask interrupts while we get this started. */
	local_irq_save(flags);

	/* Make sure the channel has at least one buffer in the queue. */
	if (dma->new_buffers < 1) {
		/* Unmask irqs */
		local_irq_restore(flags);

		pr_debug("DMA Start: Channel starved.\n");

		/*
		 * This channel does not have enough buffers queued up,
		 * so enter the pause by starvation state.
		 */
		dma->xfer_enable = 1;
		dma->pause = 1;

		/* Success. */
		return 0;
	}

	/* Clear any pending interrupts. */
	__raw_writel(0, base + EP93XX_DMA_M2M_INTERRUPT_OFFSET);

	/*
	 * Set up one or both buffer descriptors with values from the next one or
	 * two buffers in the queue.  By default disable the next frame buffer
	 * interrupt on the channel.
	 */
	ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
	ctl &= ~EP93XX_DMA_M2M_CONTROL_NFBINTEN;
	__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);

	/* enable the done interrupt. */
	ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
	ctl |= EP93XX_DMA_M2M_CONTROL_DONEINTEN;
	__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);

	/* Update the dma channel instance transfer state. */
	dma->xfer_enable = 1;
	dma->pause = 0;

	/*
	 * Program up the first buffer descriptor with a source and destination
	 * and a byte count.
	 */
	__raw_writel(dma->buffer_queue[dma->current_buffer].source,
		     base + EP93XX_DMA_M2M_SAR_BASE0_OFFSET);

	__raw_writel(dma->buffer_queue[dma->current_buffer].dest,
		     base + EP93XX_DMA_M2M_DAR_BASE0_OFFSET);

	__raw_writel(dma->buffer_queue[dma->current_buffer].size,
		     base + EP93XX_DMA_M2M_BCR0_OFFSET);

	/* Decrement the new buffers counter. */
	dma->new_buffers--;

	/*
	 * Set up the second buffer descriptor with a second buffer if we have
	 * a second buffer.
	 */
	if (dma->new_buffers) {
		__raw_writel(dma->buffer_queue[(dma->current_buffer + 1) %
					       MAX_EP93XX_DMA_BUFFERS].source,
			     base + EP93XX_DMA_M2M_SAR_BASE1_OFFSET);

		__raw_writel(dma->buffer_queue[(dma->current_buffer + 1) %
					       MAX_EP93XX_DMA_BUFFERS].dest,
			     base + EP93XX_DMA_M2M_DAR_BASE1_OFFSET);

		__raw_writel(dma->buffer_queue[(dma->current_buffer + 1) %
					       MAX_EP93XX_DMA_BUFFERS].size,
			     base + EP93XX_DMA_M2M_BCR1_OFFSET);

		ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		ctl |= EP93XX_DMA_M2M_CONTROL_NFBINTEN;
		__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);

		dma->new_buffers--;
	}

	/* Now we enable the channel.  This initiates the transfer. */
	ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
	ctl |= EP93XX_DMA_M2M_CONTROL_ENABLE;
	__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);
	__raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);

	/*
	 * If this is a memory to memory transfer, we need to s/w trigger the
	 * transfer by setting the start bit within the control register.
	 */
	if (dma->device == DMA_MEMORY) {
		ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		ctl |= EP93XX_DMA_M2M_CONTROL_START;
		__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);
	}

	pr_debug("DMA - It's been started!!\n");
	dump_m2m_register(base);

	/* Unmask irqs */
	local_irq_restore(flags);

	/* Success. */
	return 0;
}

/*****************************************************************************
 *
 * DMA interface functions
 *
 ****************************************************************************/

/*****************************************************************************
 *
 * int dma_init(int handle, unsigned int flags_m2p, unsigned int flags_m2m,
 *			   dma_callback callback, unsigned int user_data)
 *
 * Description: Configure the DMA channel and install a callback function.
 *
 * handle:	Handle unique the each instance of the dma interface, used
 *		to verify this call.
 * flags_m2p:  	Flags used to configure an M2P/P2M dma channel and determine
 *		if a callback function and user_data information are included
 *		in this call. This field should be NULL if handle represents
 *		an M2M channel.
 * flags_m2m:  	Flags used to configure an M2M dma channel and determine
 *		if a callback function and user_data information are included
 *		in this call. This field should be NULL if handle represents
 *		an M2P/P2M channel.
 * callback:	function pointer which is called near the end of the
 *		dma channel's irq handler.
 * user_data:   defined by the calling driver.
 *
 ****************************************************************************/
int
ep93xx_dma_config(int handle, unsigned int flags_m2p, unsigned int flags_m2m,
		  dma_callback callback, unsigned int user_data)
{
	int ch;
	ep93xx_dma_t *dma;
	unsigned long flags;
	unsigned int base, ctl;

	/* Get the DMA hw channel # from the handle. */
	ch = dma_get_channel_from_handle(handle);

	/* See if this is a valid handle. */
	if (ch < 0) {
		pr_err("DMA Config: Invalid dma handle.\n");
		return -EINVAL;
	}

	pr_debug("DMA Config \n");

	dma = &ep93xx_dma_info[ch];

	local_irq_save(flags);

	/* Check if the channel is currently transferring. */
	if (dma->xfer_enable) {
		local_irq_restore(flags);
		return -EINVAL;
	}

	/* Check if this is an m2m function. */
	if (ch >= MAX_EP93XX_DMA_M2P_CHANNELS) {
		local_irq_restore(flags);

		/* Call another function to handle m2m config. */
		return (dma_config_m2m(dma, flags_m2m, callback, user_data));
	}

	/* Setup a pointer into the dma channel's register set. */
	base = dma->reg_base;

	/* By default we enable the stall interrupt. */
	ctl = __raw_readl(base + EP93XX_DMA_M2P_CONTROL_OFFSET);
	ctl |= EP93XX_DMA_M2P_CONTROL_STALLINT;
	__raw_writel(ctl, base + EP93XX_DMA_M2P_CONTROL_OFFSET);

	/* Configure the channel for an error from the peripheral. */
	ctl = __raw_readl(base + EP93XX_DMA_M2P_CONTROL_OFFSET);
	if (flags_m2p && CHANNEL_ERROR_INT_ENABLE)
		ctl |= EP93XX_DMA_M2P_CONTROL_CHERRORINT;
	else
		ctl &= ~EP93XX_DMA_M2P_CONTROL_CHERRORINT;
	__raw_writel(ctl, base + EP93XX_DMA_M2P_CONTROL_OFFSET);

	ctl = __raw_readl(base + EP93XX_DMA_M2P_CONTROL_OFFSET);
	if (flags_m2p && CHANNEL_ABORT)
		ctl |= EP93XX_DMA_M2P_CONTROL_ABORT;
	else
		ctl &= ~EP93XX_DMA_M2P_CONTROL_ABORT;
	__raw_writel(ctl, base + EP93XX_DMA_M2P_CONTROL_OFFSET);

	ctl = __raw_readl(base + EP93XX_DMA_M2P_CONTROL_OFFSET);
	if (flags_m2p && IGNORE_CHANNEL_ERROR)
		ctl |= EP93XX_DMA_M2P_CONTROL_ICE;
	else
		ctl &= ~EP93XX_DMA_M2P_CONTROL_ICE;
	__raw_writel(ctl, base + EP93XX_DMA_M2P_CONTROL_OFFSET);

	/* Save the callback function in the dma instance for this channel. */
	dma->callback = callback;

	/* Save the user data in the the dma instance for this channel. */
	dma->user_data = user_data;

	/*
	 * Put the dma instance into the pause state by setting the
	 * pause bit to true.
	 */
	dma->pause = 1;

	local_irq_restore(flags);

	/* Success. */
	return 0;
}

/*****************************************************************************
 *
 * int dma_start(int handle, unsigned int channels, unsigned int * handles)
 *
 * Description: Initiate a transfer on up to 3 channels.
 *
 * handle:	handle for the channel to initiate transfer on.
 * channels:	number of channels to initiate transfers on.
 * handles:	pointer to an array of handles, one for each channel which
 *		is to be started.
 *
 ****************************************************************************/
int
ep93xx_dma_start(int handle, unsigned int channels, unsigned int *handles)
{
	ep93xx_dma_t *dma[3];
	unsigned int base[3];
	unsigned int i, ctl;
	unsigned long flags;
	int ch;

	/* Get the DMA hw channel # from the handle. */
	ch = dma_get_channel_from_handle(handle);

	/* See if this is a valid handle. */
	if (ch < 0) {
		pr_err("DMA Start: Invalid dma handle.\n");
		return -EINVAL;
	}

	if (channels < 1) {
		pr_err("DMA Start: Invalid parameter.\n");
		return -EINVAL;
	}

	pr_debug("DMA Start \n");

	/* Mask off registers. */
	local_irq_save(flags);

	/* Check if this is a start multiple. */
	if (channels > 1) {
		pr_err("DMA ERROR: Start, multiple start not supported yet \n");
		return -1;
	} else {
		/* Check if this channel is already transferring. */
		if (ep93xx_dma_info[ch].xfer_enable && !ep93xx_dma_info[ch].pause) {
			pr_err("DMA Start: Invalid command for channel %d.\n", ch);

			/* Unmask irqs */
			local_irq_restore(flags);

			/* This channel is already transferring, so return an error. */
			return -EINVAL;
		}

		/* If this is an M2M channel, call a different function. */
		if (ch >= MAX_EP93XX_DMA_M2P_CHANNELS) {
			/* Unmask irqs */
			local_irq_restore(flags);

			/* Call the m2m start function.  Only start one channel. */
			return (dma_start_m2m(ch, &ep93xx_dma_info[ch]));
		}

		/* Make sure the channel has at least one buffer in the queue. */
		if (ep93xx_dma_info[ch].new_buffers < 1) {
			pr_debug("DMA Start: Channel starved.\n");

			/*
			 * This channel does not have enough buffers queued up,
			 * so enter the pause by starvation state.
			 */
			ep93xx_dma_info[ch].xfer_enable = 1;
			ep93xx_dma_info[ch].pause = 1;

			/* Unmask irqs */
			local_irq_restore(flags);

			/* Success. */
			return 0;
		}

		/* Set up a dma instance pointer for this dma channel. */
		dma[0] = &ep93xx_dma_info[ch];

		/* Set up a pointer to the register set for this channel. */
		base[0] = dma[0]->reg_base;
	}

	/*
	 * Setup both MAXCNT registers with values from the next two buffers
	 * in the queue, and enable the next frame buffer interrupt on the channel.
	 */
	for (i = 0; i < channels; i++) {
		/* Check if we need to restore a paused transfer. */
		if (dma[i]->pause_buf.buf_id != -1)
			__raw_writel(dma[i]->pause_buf.size,
				     base[i] + EP93XX_DMA_M2P_MAXCNT0_OFFSET);
		else
			__raw_writel(dma[i]->buffer_queue[dma[i]->current_buffer].size,
				     base[i] + EP93XX_DMA_M2P_MAXCNT0_OFFSET);
	}

	for (i = 0; i < channels; i++) {
		/* Enable the specified dma channels. */
		ctl = __raw_readl(base[i] + EP93XX_DMA_M2P_CONTROL_OFFSET);
		ctl |= EP93XX_DMA_M2P_CONTROL_ENABLE;
		__raw_writel(ctl, base[i] + EP93XX_DMA_M2P_CONTROL_OFFSET);

		/* Update the dma channel instance transfer state. */
		dma[i]->xfer_enable = 1;
		dma[i]->pause = 0;
	}

	/*
	 * Program up the BASE0 registers for all specified channels, this
	 * will initiate transfers on all specified channels.
	 */
	for (i = 0; i < channels; i++)
		/* Check if we need to restore a paused transfer. */
		if (dma[i]->pause_buf.buf_id != -1) {
			__raw_writel(dma[i]->pause_buf.source,
				     base[i] + EP93XX_DMA_M2P_BASE0_OFFSET);

			/* Set the pause buffer to NULL */
			dma[i]->pause_buf.buf_id = -1;
			dma[i]->pause_buf.size = 0;
		} else if(dma[i]->new_buffers){
			__raw_writel(dma[i]->buffer_queue[dma[i]->current_buffer].source,
				     base[i] + EP93XX_DMA_M2P_BASE0_OFFSET);
			dma[i]->new_buffers--;
		}

	/*
	 * Before restoring irqs setup the second MAXCNT/BASE
	 * register with a second buffer.
	 */
	for (i = 0; i < channels; i++) {
		if (dma[i]->new_buffers) {
			/* By default we enable the next frame buffer interrupt. */
			ctl = __raw_readl(base[i] + EP93XX_DMA_M2P_CONTROL_OFFSET);
			ctl |= EP93XX_DMA_M2P_CONTROL_NFBINT;
			__raw_writel(ctl, base[i] + EP93XX_DMA_M2P_CONTROL_OFFSET);

			__raw_writel(dma[i]->buffer_queue[(dma[i]->current_buffer + 1) %
							  MAX_EP93XX_DMA_BUFFERS].size,
				     base[i] + EP93XX_DMA_M2P_MAXCNT1_OFFSET);

			__raw_writel(dma[i]->buffer_queue[(dma[i]->current_buffer + 1) %
							  MAX_EP93XX_DMA_BUFFERS].source,
				     base[i] + EP93XX_DMA_M2P_BASE1_OFFSET);
			dma[i]->new_buffers--;
		}
	}

	/* Unmask irqs */
	local_irq_restore(flags);

	/* Success. */
	return 0;
}

/*****************************************************************************
 *
 * int ep93xx_dma_add_buffer(int handle, unsigned int * address,
 *				unsigned int size, unsigned int last)
 *
 * Description: Add a buffer entry to the DMA buffer queue.
 *
 * handle:	handle for the channel to add this buffer to.
 * address:	Pointer to an integer which is the start address of the
 *		buffer which is to be added to the queue.
 * size:	size of the buffer in bytes.
 * last:	1 if this is the last buffer in this stream, 0 otherwise.
 *
 ****************************************************************************/
int
ep93xx_dma_add_buffer(int handle, unsigned int source, unsigned int dest,
		      unsigned int size, unsigned int last, unsigned int buf_id)
{
	unsigned long flags;
	ep93xx_dma_t *dma;
	int ch;

	/* Get the DMA hw channel # from the handle. */
	ch = dma_get_channel_from_handle(handle);

	/* See if this is a valid handle. */
	if (ch < 0) {
		pr_err("DMA Add Buffer: Invalid dma handle.\n");
		return -EINVAL;
	}

	/* Get a pointer to the dma instance. */
	dma = &ep93xx_dma_info[ch];

	/* Mask interrupts and hold on to the original state. */
	local_irq_save(flags);

	/*
	 * If the buffer queue is full, last_buffer is the same as current_buffer and
	 * we're not tranfering, or last_buffer is pointing to a used buffer,
	 * then exit.
	 * TODO: do I need to do any more checks?
	 */
	if (dma->total_buffers >= MAX_EP93XX_DMA_BUFFERS) {
		/* Restore the state of the irqs */
		local_irq_restore(flags);

		/* Fail. */
		return -1;
	}

	/* Add this buffer to the queue */
	dma->buffer_queue[dma->last_buffer].source = source;
	dma->buffer_queue[dma->last_buffer].dest = dest;
	dma->buffer_queue[dma->last_buffer].size = size;
	dma->buffer_queue[dma->last_buffer].last = last;
	dma->buffer_queue[dma->last_buffer].buf_id = buf_id;

	/* Reset the used field of the buffer structure. */
	dma->buffer_queue[dma->last_buffer].used = 0;

	/* Increment the End Item Pointer. */
	dma->last_buffer = (dma->last_buffer + 1) % MAX_EP93XX_DMA_BUFFERS;

	/* Increment the new buffers counter and the total buffers counter */
	dma->new_buffers++;
	dma->total_buffers++;

	/* restore the interrupt state. */
	local_irq_restore(flags);

	/* Check if the channel was starved into a stopped state. */
	if (dma->pause && dma->xfer_enable) {
		if (dma->new_buffers >= 1) {
			pr_debug("DMA - calling start from add after starve. \n");

			/*
			 * The channel was starved into a stopped state, and we've got
			 * 2 new buffers, so start tranferring again.
			 */
			ep93xx_dma_start(handle, 1, 0);
		}
	}

	/* Success. */
	return 0;
}

/*****************************************************************************
 *
 * int ep93xx_dma_remove_buffer(int handle, unsigned int * address,
 *				unsigned int * size)
 *
 * Description: Remove a buffer entry from the DMA buffer queue. If
 *		buffer was removed successfully, return 0, otherwise
 *		return -1.
 *
 * handle:	handle for the channel to remove a buffer from.
 * address:	Pointer to an integer which is filled in with the start
 *		address of the removed buffer.
 * size:	Pointer to an integer which is filled in with the size in
 *		bytes of the removed buffer.
 *
 ****************************************************************************/
int
ep93xx_dma_remove_buffer(int handle, unsigned int * buf_id)
{
	unsigned int test;
	unsigned int loop;
	int ret = -1;
	unsigned long flags;
	ep93xx_dma_t *dma;
	int ch;

	/* Get the DMA hw channel # from the handle. */
	ch = dma_get_channel_from_handle(handle);

	/* See if this is a valid handle. */
	if (ch < 0) {
		pr_err("DMA Remove Buffer: Invalid dma handle.\n");
		return -EINVAL;
	}

	dma = &ep93xx_dma_info[ch];

	/* Mask interrupts and hold on to the original state. */
	local_irq_save(flags);

	/* Make sure there are used buffers to be returned. */
	if (dma->used_buffers) {
		test = dma->last_buffer;

		for (loop = 0; loop < MAX_EP93XX_DMA_BUFFERS; loop++) {
			if (dma->buffer_queue[test].used &&
			    (dma->buffer_queue[test].buf_id != -1)) {
				/*
				 * This is a used buffer, fill in the buf_id pointer
				 * with the buf_id for this buffer.
				 */
				*buf_id = dma->buffer_queue[test].buf_id;

				/* Reset this buffer structure */
				dma->buffer_queue[test].buf_id = -1;

				/*
				 * Decrement the used buffer counter, and the total
				 * buffer counter.
				 */
				dma->used_buffers--;
				dma->total_buffers--;

				/*
				 * Successful removal of a buffer, so set the return
				 * value to 0, then exit this loop.
				 */
				ret = 0;
				break;
			}

			/* This buffer isn't used, let's see if the next one is. */
			test = (test + 1) % MAX_EP93XX_DMA_BUFFERS;
		}
	}

	/* Restore interrupts. */
	local_irq_restore(flags);

	/* Success. */
	return ret;
}

/*****************************************************************************
 *
 * int ep93xx_dma_pause(int handle, unsigned int channels,
 *			unsigned int * handles)
 *
 * Description: Disable any ongoing transfer for the given channel, retaining
 *		the state of the current buffer transaction so that upon
 *		resume, the dma will continue where it left off.
 *
 * handle:	Handle for the channel to be paused.  If this is a pause for
 *		for multiple channels, handle is a valid handle for one of
 *		the channels to be paused.
 * channels:	number of channel to pause transfers on.
 * handles:	Pointer to an array of handles, one for each channel which
 *		to be paused.  If this pause is intended only for one
 *		channel, this field should be set to NULL.
 *
 ****************************************************************************/
int
ep93xx_dma_pause(int handle, unsigned int channels, unsigned int * handles)
{
	unsigned long flags;
	ep93xx_dma_t *dma;
	int ch;

	pr_debug("ep93xx_dma_pause\n");

	/* Mask interrupts and hold on to the original state. */
	local_irq_save(flags);

	/* Get the DMA hw channel # from the handle. */
	ch = dma_get_channel_from_handle(handle);

	/* See if this is a valid handle. */
	if (ch < 0) {
		/* restore interrupts. */
		local_irq_restore(flags);

		pr_err("DMA Pause: Invalid dma handle.\n");

		/* Fail. */
		return -EINVAL;
	}

	pr_debug("DMA %d: pause \n", ch);

	/* Set up a pointer to the dma instance data. */
	dma = &ep93xx_dma_info[ch];

	/* Check if we're already paused. */
	if (dma->pause) {
		/* We're paused, but are we stopped? */
		if (dma->xfer_enable)
			/* Put the channel in the stopped state. */
			dma->xfer_enable = 0;

		pr_debug("DMA Pause - already paused.");
	} else {
		/* Put the channel into the stopped state. */
		dma->xfer_enable = 0;
		dma->pause = 1;
	}

	/* restore interrupts. */
	local_irq_restore(flags);

	/* Already paused, so exit. */
	return 0;
}

/*****************************************************************************
 *
 * void ep93xx_dma_flush(int handle)
 *
 * Description: Flushes all queued buffers and transfers in progress
 *		for the given channel.  Return the buffer entries
 *		to the calling function.
 *
 * handle:	handle for the channel for which the flush is intended.
 *
 ****************************************************************************/
int
ep93xx_dma_flush(int handle)
{
	unsigned int i;
	unsigned long flags;
	ep93xx_dma_t *dma;
	int ch;
	unsigned int base, ctl;

	/* Get the DMA hw channel # from the handle. */
	ch = dma_get_channel_from_handle(handle);

	/* See if this is a valid handle. */
	if (ch < 0) {
		pr_err("DMA Flush: Invalid dma handle.\n");
		return -EINVAL;
	}

	pr_debug("DMA %d: flush \n", ch);

	/* Set up a pointer to the dma instance data for this channel */
	dma = &ep93xx_dma_info[ch];

	/* Mask interrupts and hold on to the original state. */
	local_irq_save(flags);

	/* Disable the dma channel */
	if (ch < MAX_EP93XX_DMA_M2P_CHANNELS) {
		/* M2P channel */
		ctl = __raw_readl(dma->reg_base + EP93XX_DMA_M2P_CONTROL_OFFSET);
		ctl &= ~EP93XX_DMA_M2P_CONTROL_ENABLE;
		__raw_writel(ctl, dma->reg_base + EP93XX_DMA_M2P_CONTROL_OFFSET);
	} else {
		/* M2M channel */
		ctl = __raw_readl(dma->reg_base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		ctl &= ~EP93XX_DMA_M2M_CONTROL_ENABLE;
		__raw_writel(ctl, dma->reg_base + EP93XX_DMA_M2M_CONTROL_OFFSET);
	}

	for (i = 0; i < MAX_EP93XX_DMA_BUFFERS; i++) {
		dma->buffer_queue[i].buf_id = -1;
		dma->buffer_queue[i].last = 0;
	}

	/* Set the Current and Last item to zero. */
	dma->current_buffer = 0;
	dma->last_buffer = 0;

	/* Reset the Buffer counters */
	dma->used_buffers = 0;
	dma->new_buffers = 0;
	dma->total_buffers = 0;

	/* reset the Total bytes counter. */
	dma->total_bytes = 0;

	/* Reset the paused buffer. */
	dma->pause_buf.last = 0;
	dma->pause_buf.buf_id = -1;

	base = ep93xx_dma_info[ch].reg_base;

	/* restore interrupts. */
	local_irq_restore(flags);

	/* Success. */
	return 0;
}

/*****************************************************************************
 *
 * int ep93xx_dma_queue_full(int handle)
 *
 * Description: Query to determine if the DMA queue of buffers for
 *			  a given channel is full.
 *			  0 = queue is full
 *			  1 = queue is not full
 *
 * handle:	 handle for the channel to query.
 *
 ****************************************************************************/
int
ep93xx_dma_queue_full(int handle)
{
	int ret = 0;
	unsigned long flags;
	int ch;

	/* Get the DMA hw channel # from the handle. */
	ch = dma_get_channel_from_handle(handle);

	/* See if this is a valid handle. */
	if (ch < 0) {
		pr_err("DMA Queue Full: Invalid dma handle.\n");
		return -EINVAL;
	}

	pr_debug("DMA %d: queue full \n", ch);

	/* Mask interrupts and hold on to the original state. */
	local_irq_save(flags);

	/*
	 * If the last item is equal to the used item then
	 * the queue is full.
	 */
	if (ep93xx_dma_info[ch].total_buffers < MAX_EP93XX_DMA_BUFFERS)
		ret =  0;
	else
		ret = 1;

	/* restore interrupts. */
	local_irq_restore(flags);

	return ret;
}

/*****************************************************************************
 *
 * int ep93xx_dma_get_position()
 *
 * Description:  Takes two integer pointers and fills them with the start
 *               and current address of the buffer currently transferring
 *               on the specified DMA channel.
 *
 * handle         handle for the channel to query.
 * *buf_id        buffer id for the current buffer transferring on the
 *                dma channel.
 * *total         total bytes transferred on the channel.  Only counts  
 *                whole buffers transferred.
 * *current_frac  number of bytes transferred so far in the current buffer.
 ****************************************************************************/
int
ep93xx_dma_get_position(int handle, unsigned int *buf_id,
                        unsigned int *total, unsigned int *current_frac)
{
	int ch;
	ep93xx_dma_t *dma;
	unsigned int id1, total1, cur_frac1, id2, total2;
	unsigned int status, next, next_state, base = 0;
	unsigned int pause1, pause2;

	/*
	 * Get the DMA hw channel # from the handle.  See if this is a 
	 * valid handle.
	 */
	ch = dma_get_channel_from_handle(handle);
	if (ch < 0) {
		pr_err("DMA Get Position: Invalid dma handle.\n");
		return -EINVAL;
	}

	dma = &ep93xx_dma_info[ch];

	/*
	 * If DMA moves to a new buffer in the middle of us grabbing the 
	 * buffer info, then do it over again.
	 */
	do {
		id1 = dma->buffer_queue[dma->current_buffer].buf_id;
		total1 = dma->total_bytes;
		pause1 = dma->pause;

		if (ch < MAX_EP93XX_DMA_M2P_CHANNELS) {
			/* M2P */
			base = dma->reg_base;

			status = __raw_readl(base + EP93XX_DMA_M2P_STATUS_OFFSET);

			next = ((status & EP93XX_DMA_M2P_STATUS_NEXTBUFFER) != 0);

			next_state = (((status & EP93XX_DMA_M2P_STATUS_CURRENT_MASK) >>
				       EP93XX_DMA_M2P_STATUS_CURRENT_SHIFT) == 
				      EP93XX_DMA_M2P_STATUS_NEXTBUFFER);
			
			if (next ^ next_state)
				cur_frac1 = __raw_readl(base+EP93XX_DMA_M2P_CURRENT1_OFFSET) -
					__raw_readl(base+EP93XX_DMA_M2P_BASE1_OFFSET);
			else
				cur_frac1 = __raw_readl(base+EP93XX_DMA_M2P_CURRENT0_OFFSET) -
					__raw_readl(base+EP93XX_DMA_M2P_BASE0_OFFSET);
		} else {
			/* M2M - TODO implement this for M2M*/
			cur_frac1 = 0;
		}
		
		id2 = dma->buffer_queue[dma->current_buffer].buf_id;
		total2 = dma->total_bytes;
		pause2 = dma->pause;

	} while ((id1 != id2) || (total1 != total2) || (pause1 != pause2));

	if (pause1)
		cur_frac1 = 0;

	if (buf_id)
		*buf_id = id1;

	if (total)
		*total = total1;

	if (current_frac)
		*current_frac = cur_frac1;
	
	/* Success. */
	return 0;
}

/*****************************************************************************
 *
 * int ep93xx_dma_get_total(int handle)
 *
 * Description:	Returns the total number of bytes transferred on the
 *		specified channel since the channel was requested.
 *
 * handle:	handle for the channel to query.
 *
 ****************************************************************************/
int
ep93xx_dma_get_total(int handle)
{
	int ch;

	/* Get the DMA hw channel # from the handle. */
	ch = dma_get_channel_from_handle(handle);

	/* See if this is a valid handle. */
	if (ch < 0) {
		pr_err("DMA Get Total: Invalid dma handle.\n");
		return -EINVAL;
	}

	pr_debug("DMA %d: total: %d \n", ch, ep93xx_dma_info[ch].total_bytes);

	/*
	 * Return the total number of bytes transferred on this channel since
	 * it was requested.
	 */
	return ep93xx_dma_info[ch].total_bytes;
}

/*****************************************************************************
 *
 * int ep93xx_dma_is_done(int handle)
 *
 * Description:	Determines if the specified channel is done
 *		transferring the requested data.
 *
 * handle:	handle for the channel to query.
 *
 ****************************************************************************/
int
ep93xx_dma_is_done(int handle)
{
	ep93xx_dma_t *dma;
	int ch;

	/* Get the DMA hw channel # from the handle. */
	ch = dma_get_channel_from_handle(handle);

	/* See if this is a valid handle. */
	if (ch < 0) {
		pr_err("ep93xx_dma_is_done: Invalid dma handle.\n");
		return -EINVAL;
	}

	/* Get a pointer to the DMA channel state structure. */
	dma = &ep93xx_dma_info[ch];

	/* See if there are any buffers remaining to be provided to the HW. */
	if (dma->new_buffers)
		return 0;

	/* See if this is a M2P or M2M channel. */
	if (ch < MAX_EP93XX_DMA_M2P_CHANNELS) {
		/*
		 * If the bytes remaining register of the HW is not zero, then
		 * there is more work to be done.
		 */
		if (__raw_readl(dma->reg_base + EP93XX_DMA_M2P_REMAIN_OFFSET) != 0)
			return 0;
	} else {
		/*
		 * If either byte count register in the HW is not zero, then there
		 * is more work to be done.
		 */
		if ((__raw_readl(dma->reg_base + EP93XX_DMA_M2M_BCR0_OFFSET) != 0) ||
		    (__raw_readl(dma->reg_base + EP93XX_DMA_M2M_BCR1_OFFSET) != 0))
			return 0;
	}

	/* The DMA is complete. */
	return 1;
}

/*****************************************************************************
 * ep93xx_dma_request
 *
 * Description: This function will allocate a DMA channel for a particular
 * hardware peripheral.  Before initiating a transfer on the allocated
 * channel, the channel must be set up and buffers have to queued up.
 *
 * handle:	pointer to an integer which is filled in with a unique
 *		handle for this instance of the dma interface.
 * device_id:   string with the device name, primarily used by /proc.
 * device:	hardware device ID for which the requested dma channel will
 *		transfer data.
 *
 ****************************************************************************/
int
ep93xx_dma_request(int *handle, const char *device_id,
		   ep93xx_dma_dev_t device)
{
	ep93xx_dma_t *dma = NULL;
	int ch;
	unsigned int error = 0;
	unsigned int i;
	unsigned int base;

	/* Check if the device requesting a DMA channel is a valid device. */
	if ((device >= UNDEF) || (device < 0))
		return -ENODEV;

	/*
	 * We've got a valid hardware device requesting a DMA channel.
	 * Now check if the device should open an M2P or M2M channel
	 */
	if (device < 20)
		ch = dma_open_m2p(device);
	else
		ch = dma_open_m2m(device);

	/* Check if we successfully opened a DMA channel */
	if (ch < 0) {
		pr_err("%s: Could not open dma channel for this device.\n",
		       device_id);
		return -EBUSY;
	}

	dma = &ep93xx_dma_info[ch];

	/* Request the appropriate IRQ for the specified channel */
	if (ch < MAX_EP93XX_DMA_M2P_CHANNELS)
		error = request_irq(dma->irq, dma_m2p_irq_handler,
				    IRQF_DISABLED, device_id, (void *)dma);
	else
		error = request_irq(dma->irq, &dma_m2m_irq_handler,
				    IRQF_DISABLED, device_id, (void *)dma);

	/* Check for any errors during the irq request */
	if (error) {
		pr_err("%s: unable to request IRQ %d for DMA channel\n",
		       device_id, dma->irq);
		return error;
	}

	/*
	 * Generate a valid handle and exit.
	 *
	 * Increment the last valid handle.
	 * Check for wraparound (unlikely, but we like to be complete).
	 */
	dma->last_valid_handle++;

	if ((dma->last_valid_handle & DMA_HANDLE_SPECIFIER_MASK) != (ch << 28))
		dma->last_valid_handle = (ch << 28) + 1;

	/*
	 * Fill in the handle pointer with a valid handle for
	 * this dma channel instance.
	 */
	*handle = dma->last_valid_handle;

	pr_debug("Handle for channel %d: 0x%x\n", ch, *handle);

	/* Save the device ID and device name. */
	dma->device = device;
	dma->device_id = device_id;

	/* Init all fields within the dma instance. */
	for (i = 0; i < MAX_EP93XX_DMA_BUFFERS; i++)
		dma->buffer_queue[i].buf_id = -1;

	/* Initialize all buffer queue variables. */
	dma->current_buffer = 0;
	dma->last_buffer = 0;

	dma->new_buffers = 0;
	dma->used_buffers = 0;
	dma->total_buffers = 0;

	/* Initialize the total bytes variable */
	dma->total_bytes = 0;

	/* Initialize the transfer and pause state variables to 0. */
	dma->xfer_enable = 0;

	dma->pause = 0;

	/* Initialize the pause buffer structure. */
	dma->pause_buf.buf_id = -1;

	/* Initialize the callback function and user data fields. */
	dma->callback = NULL;

	/*
	 * User data used as a parameter for the Callback function.  The user
	 * sets up the data and sends it with the callback function.
	 */
	dma->user_data = 0;

	base = ep93xx_dma_info[ch].reg_base;

	/* Debugging message. */
	pr_debug("Successfully requested dma channel %d\n", ch);
	dump_m2p_register(base);
	dump_dma_info(dma);

	pr_debug("VIC0IRQSTATUS - 0x%x, VIC0INTENABLE - 0x%x \n",
		 __raw_readl(EP93XX_VIC1_BASE + VIC_IRQ_STATUS),
		 __raw_readl(EP93XX_VIC1_BASE + VIC_INT_ENABLE));

	/* Success. */
	return 0;
}

/*****************************************************************************
 *
 * ep93xx_dma_free
 *
 * Description: This function will free the dma channel for future requests.
 *
 * handle:	 handle for the channel to be freed.
 *
 ****************************************************************************/
int
ep93xx_dma_free(int handle)
{
	ep93xx_dma_t *dma;
	unsigned int base, ctl;
	int ch;

	/* Get the DMA hw channel # from the handle. */
	ch = dma_get_channel_from_handle(handle);

	/* See if this is a valid handle. */
	if (ch < 0) {
		pr_err("DMA Free: Invalid dma handle.\n");
		return -EINVAL;
	}

	/* Get a pointer to the dma instance. */
	dma = &ep93xx_dma_info[ch];

	/* Disable the dma channel */
	if (ch < MAX_EP93XX_DMA_M2P_CHANNELS) {
		/* M2P channel */
		base = dma->reg_base;

		ctl = __raw_readl(base + EP93XX_DMA_M2P_CONTROL_OFFSET);
		ctl &= ~EP93XX_DMA_M2P_CONTROL_ENABLE;
		__raw_writel(ctl, base + EP93XX_DMA_M2P_CONTROL_OFFSET);
	} else {
		/* M2M channel */
		base = dma->reg_base;

		ctl = __raw_readl(base + EP93XX_DMA_M2M_CONTROL_OFFSET);
		ctl &= ~EP93XX_DMA_M2M_CONTROL_ENABLE;
		__raw_writel(ctl, base + EP93XX_DMA_M2M_CONTROL_OFFSET);
	}

	/* Free the interrupt servicing this dma channel */
	free_irq(dma->irq, dma);

	/* Decrement the reference count for this instance of the dma interface */
	dma->ref_count--;

	/*
	 * Set the transfer and pause state variables to 0
	 * (unititialized state).
	 */
	dma->xfer_enable = 0;
	dma->pause = 0;

	/* Debugging message. */
	pr_debug("Successfully freed dma channel %d\n", ch);

	/* Success. */
	return 0;
}

/*****************************************************************************
 *
 * ep93xx_dma_init(void)
 *
 * Description: This function is called during system initialization to
 * setup the interrupt number and register set base address for each DMA
 * channel.
 *
 ****************************************************************************/
static int __init
ep93xx_dma_init(void)
{
	int ch;

	/* Init some values in each dma instance. */
	for (ch = 0; ch < MAX_EP93XX_DMA_CHANNELS; ch++) {
		/* IRQ for the specified dma channel. */
		ep93xx_dma_info[ch].irq = IRQ_EP93XX_DMAM2P0 + ch;

		/* Initial value of the dma channel handle. */
		ep93xx_dma_info[ch].last_valid_handle = ch << 28;

		/*
		 * Give the instance a pointer to the dma channel register
		 * base.
		 */
		if (ch < MAX_EP93XX_DMA_M2P_CHANNELS)
			ep93xx_dma_info[ch].reg_base = ep93xx_dma_m2p_base[ch];
		else
			ep93xx_dma_info[ch].reg_base = ep93xx_dma_m2m_base[ch - 10];

		/* Initialize the reference count for this channel. */
		ep93xx_dma_info[ch].ref_count = 0;
	}

	pr_debug("DMA Interface intitialization complete\n");

	/* Success */
	return 0;
}

arch_initcall(ep93xx_dma_init);

EXPORT_SYMBOL(ep93xx_dma_free);
EXPORT_SYMBOL(ep93xx_dma_request);
EXPORT_SYMBOL(ep93xx_dma_flush);
EXPORT_SYMBOL(ep93xx_dma_pause);
EXPORT_SYMBOL(ep93xx_dma_remove_buffer);
EXPORT_SYMBOL(ep93xx_dma_add_buffer);
EXPORT_SYMBOL(ep93xx_dma_start);
EXPORT_SYMBOL(ep93xx_dma_config);

